/**
 * OPEN-XCHANGE legal information
 *
 * All intellectual property rights in the Software are protected by
 * international copyright laws.
 *
 *
 * In some countries OX, OX Open-Xchange, open xchange and OXtender
 * as well as the corresponding Logos OX Open-Xchange and OX are registered
 * trademarks of the Open-Xchange, Inc. group of companies.
 * The use of the Logos is not covered by the GNU General Public License.
 * Instead, you are allowed to use these Logos according to the terms and
 * conditions of the Creative Commons License, Version 2.5, Attribution,
 * Non-commercial, ShareAlike, and the interpretation of the term
 * Non-commercial applicable to the aforementioned license is published
 * on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 * Please make sure that third-party modules and libraries are used
 * according to their respective licenses.
 *
 * Any modifications to this package must retain all copyright notices
 * of the original copyright holder(s) for the original code used.
 *
 * After any such modifications, the original and derivative code shall remain
 * under the copyright of the copyright holder(s) and/or original author(s)per
 * the Attribution and Assignment Agreement that can be located at
 * http://www.open-xchange.com/EN/developer/. The contributing author shall be
 * given Attribution for the derivative code and a license granting use.
 *
 *  Copyright (C) 2004-2014 Open-Xchange, Inc.
 *  Mail: info@open-xchange.com
 *
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, Version 2 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc., 59
 *  Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package com.openxchange.office_communication.office.jms.messages;

import org.apache.commons.lang3.Validate;

import com.openxchange.office_communication.jms.core.plain.beans.JMSHeader;
import com.openxchange.office_communication.jms.core.plain.beans.JMSMessageBean;

//=============================================================================
public class DocumentMessageBean extends JMSMessageBean
{
	//-------------------------------------------------------------------------
	public static final String HEADER_NAMESPACE   = JMSMessageBean.HEADER_NAMESPACE+JMSHeader.SEPARATOR+"document";
	
	//-------------------------------------------------------------------------
	public static final String HEADER_DOCUMENT_ID = DocumentMessageBean.HEADER_NAMESPACE+JMSHeader.SEPARATOR+"id";
	public static final String HEADER_DOCUMENT_OP = DocumentMessageBean.HEADER_NAMESPACE+JMSHeader.SEPARATOR+"op";

    //-------------------------------------------------------------------------
	public DocumentMessageBean ()
		throws Exception
	{
		//setType(DocumentMessageBean.class);
	}

	//-------------------------------------------------------------------------
	public static DocumentMessageBean newMessage (final String             sDocID,
											      final EDocumentOperation eOp   )
		throws Exception
	{
		final DocumentMessageBean aMsg = new DocumentMessageBean ();
		aMsg.setDocumentID(sDocID);
		aMsg.setOperation (eOp   );
		return aMsg;
	}

	//-------------------------------------------------------------------------
	@Override
	public void validate ()
		throws Exception
	{
		Validate.notEmpty(m_sDocID              ,
				          "Invalid document id.");

		Validate.notNull (m_eOp                        ,
 		        		  "Invalid document operation.");
	}

	//-------------------------------------------------------------------------
	public void setDocumentID (final String sID)
	    throws Exception
	{
		m_sDocID = sID;
	}
	
	//-------------------------------------------------------------------------
	public String getDocumentID ()
	    throws Exception
	{
		return m_sDocID;
	}

	//-------------------------------------------------------------------------
	public void setOperation (final EDocumentOperation eOp)
		throws Exception
	{
		m_eOp = eOp;
	}
	
	//-------------------------------------------------------------------------
	public EDocumentOperation getOperation ()
		throws Exception
	{
		return m_eOp;
	}

	//-------------------------------------------------------------------------
	@Override
	public String toString ()
	{
		final StringBuffer sStr = new StringBuffer (256);
		sStr.append (super.toString ());
		sStr.append ("[doc-id="       );
		sStr.append (m_sDocID         );
		sStr.append (",op="           );
		sStr.append (m_eOp            );
		sStr.append ("]"              );
		return sStr.toString ();
	}
	
	//-------------------------------------------------------------------------
	@JMSHeader(name=DocumentMessageBean.HEADER_DOCUMENT_ID)
	private String m_sDocID = null;

	//-------------------------------------------------------------------------
	@JMSHeader(name=DocumentMessageBean.HEADER_DOCUMENT_OP)
	private EDocumentOperation m_eOp = null;
}
