/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2012 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.ooxml.xlsx.operations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.docx4j.openpackaging.packages.SpreadsheetMLPackage;
import org.docx4j.openpackaging.parts.Part;
import org.docx4j.openpackaging.parts.SpreadsheetML.WorkbookPart;
import org.docx4j.openpackaging.parts.SpreadsheetML.WorksheetPart;
import org.docx4j.openpackaging.parts.relationships.RelationshipsPart;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.xlsx4j.sml.BookViews;
import org.xlsx4j.sml.CTBookView;
import org.xlsx4j.sml.CTCellFormula;
import org.xlsx4j.sml.CTCellStyle;
import org.xlsx4j.sml.CTCellStyleXfs;
import org.xlsx4j.sml.CTCellStyles;
import org.xlsx4j.sml.CTCellXfs;
import org.xlsx4j.sml.CTDefinedName;
import org.xlsx4j.sml.CTDrawing;
import org.xlsx4j.sml.CTHyperlink;
import org.xlsx4j.sml.CTHyperlinks;
import org.xlsx4j.sml.CTRst;
import org.xlsx4j.sml.CTSheetFormatPr;
import org.xlsx4j.sml.CTStylesheet;
import org.xlsx4j.sml.CTXf;
import org.xlsx4j.sml.Cell;
import org.xlsx4j.sml.Col;
import org.xlsx4j.sml.Cols;
import org.xlsx4j.sml.DefinedNames;
import org.xlsx4j.sml.IDataValidation;
import org.xlsx4j.sml.IDataValidations;
import org.xlsx4j.sml.Row;
import org.xlsx4j.sml.STCellFormulaType;
import org.xlsx4j.sml.STCellType;
import org.xlsx4j.sml.Sheet;
import org.xlsx4j.sml.SheetData;
import org.xlsx4j.sml.Sheets;
import org.xlsx4j.sml.SmlUtils;
import org.xlsx4j.sml.Workbook;
import org.xlsx4j.sml.Worksheet;

import com.openexchange.log.LogFactory;
import com.openexchange.office.FilterException;
import com.openexchange.office.FilterException.ErrorCode;
import com.openexchange.office.ooxml.tools.Commons;
import com.openexchange.office.ooxml.xlsx.OperationDocument;
import com.openexchange.office.ooxml.xlsx.tools.CellUtils;
import com.openexchange.office.ooxml.xlsx.tools.ColumnUtils;
import com.openexchange.office.ooxml.xlsx.tools.Drawings;
import com.openexchange.office.ooxml.xlsx.tools.MergeCellHelper;
import com.openexchange.office.ooxml.xlsx.tools.RowUtils;
import com.openexchange.office.ooxml.xlsx.tools.SheetUtils;
import com.openexchange.office.ooxml.xlsx.tools.ValidationUtils;
import com.openexchange.office.tools.ConfigurationHelper;
import com.openexchange.server.ServiceLookup;
import com.openexchange.session.Session;

public class CreateOperationHelper extends com.openexchange.office.ooxml.operations.CreateOperationHelper {

    private final static Log log = LogFactory.getLog(CreateOperationHelper.class);

    private final SpreadsheetMLPackage spreadsheetMLPackage;

    private final Map<Long, String> indexToStyleId = new HashMap<Long, String>();

    public CreateOperationHelper(OperationDocument _operationDocument, JSONArray mOperationsArray)
    	throws FilterException {

    	super(_operationDocument, mOperationsArray);
        spreadsheetMLPackage = _operationDocument.getPackage();

        checkComplexity();
        CTStylesheet stylesheet = getOperationDocument().getStylesheet(false);
        if(stylesheet!=null) {
            // creating styles
            final CTCellStyleXfs cellStyleXfs = stylesheet.getCellStyleXfs();
            if(cellStyleXfs!=null) {
                BuiltinIds.getUsedStyles(stylesheet, indexToStyleId, null);
            }
        }
    }

    @Override
    public OperationDocument getOperationDocument() {
        return (OperationDocument)operationDocument;
    }

    @Override
    public void createDocumentDefaults(String userLanguage) {

    }

    public void createStyleOperations()
        throws Exception {

        CTStylesheet stylesheet = getOperationDocument().getStylesheet(false);
        if(stylesheet!=null) {
            // creating styles
            final CTCellStyleXfs cellStyleXfs = stylesheet.getCellStyleXfs();
            if(cellStyleXfs!=null) {
                final List<CTXf> xfs = cellStyleXfs.getXf();
                for(int i=0;i<xfs.size();i++) {
                    final String styleId = indexToStyleId.get(new Long(i));
                    if(styleId!=null) {
                        String styleName = styleId;
                        JSONObject jsonProperties = new JSONObject();
                        Commons.mergeJsonObjectIfUsed("cell", jsonProperties, CellUtils.createCellProperties(xfs.get(i), stylesheet, true));
                        Commons.mergeJsonObjectIfUsed("character", jsonProperties, CellUtils.createCharacterProperties(xfs.get(i), stylesheet, true));
                        JSONObject attrs = jsonProperties.length()>0?jsonProperties:null;
                        String parentId = null;
                        Boolean hidden = null;
                        Integer uiPriority = null;
                        Boolean isDefault = null;

                        // check for proper UI styleName
                        CTCellStyles cellStyles = stylesheet.getCellStyles();
                        if(cellStyles!=null) {
                            for(CTCellStyle cellStyle:cellStyles.getCellStyle()) {
                                if(cellStyle.getXfId()==i) {
                                    final String name = cellStyle.getName();
                                    if(name!=null&&!name.isEmpty()) {
                                        styleName = name;
                                    }
                                    if(i==0) {
                                        isDefault = new Boolean(true);
                                    }
                                    hidden = cellStyle.isHidden();
                                    break;
                                }
                            }
                        }
                        operationsArray.put(createInsertStyleSheetOperation("cell", styleId, styleName, attrs, parentId, hidden, uiPriority, isDefault, false));
                    }
                }
            }

            // creating autostyles
            final CTCellXfs cellXfs = stylesheet.getCellXfs();
            if(cellXfs!=null) {
                final List<CTXf> cellXfsList = cellXfs.getXf();
                for(int i=0;i<cellXfsList.size();i++) {
                    final CTXf xf = cellXfsList.get(i);
                    final String styleId = "a" + Integer.toString(i);
                    String styleName = "";
                    JSONObject jsonProperties = new JSONObject();
                    Commons.mergeJsonObjectIfUsed("cell", jsonProperties, CellUtils.createCellProperties(xf, stylesheet, true));
                    Commons.mergeJsonObjectIfUsed("character", jsonProperties, CellUtils.createCharacterProperties(xf, stylesheet, true));
                    JSONObject attrs = jsonProperties.length()>0?jsonProperties:null;
                    String parentId = null;
                    final Long xfId = xf.getXfId();
                    if(xfId!=null) {
                        parentId = indexToStyleId.get(xfId);
                    }
                    if(parentId == null) {
                        parentId = "0";
                    }
                    Boolean hidden = true;
                    Integer uiPriority = null;
                    Boolean isDefault = false;
                    operationsArray.put(createInsertStyleSheetOperation("cell", styleId, styleName, attrs, parentId, hidden, uiPriority, isDefault, true));
                }
            }
        }
    }

    
    public void createOperations()
    	throws Exception {

    	final WorkbookPart workbookPart = spreadsheetMLPackage.getWorkbookPart();
	    final Workbook workbook = workbookPart.getJaxbElement();

	    createDocumentDefaults(getOperationDocument().getUserLanguage());
	    createDocumentSettings(workbook);
	    createStyleOperations();
	    CreateThemeOperations();
	    createInsertNameOperations(workbook);
		final Sheets sheets = workbook.getSheets();
	    final List<Sheet> sheetList = sheets.getSheet();
	    for(int sheetIndex = 0; sheetIndex < sheetList.size(); sheetIndex++) {
	    	final Sheet sheet = sheetList.get(sheetIndex);
	    	final Part part = getSheetPart(sheet);
	    	final String sheetType = getSheetType(part);
	        createInsertSheetOperation(sheet, sheetType, part, sheetIndex);
	        if(sheetType.equals("worksheet")) {
        		createWorksheetOperations((WorksheetPart)part, sheetIndex, false);
	        }
	    }
    }
    
    public int createPreviewOperations()
    	throws Exception {

    	final WorkbookPart workbookPart = spreadsheetMLPackage.getWorkbookPart();
	    final Workbook workbook = workbookPart.getJaxbElement();

	    createDocumentDefaults(getOperationDocument().getUserLanguage());
	    final int activeSheetIndex = createDocumentSettings(workbook);
	    createStyleOperations();
	    CreateThemeOperations();
	    createInsertNameOperations(workbook);
		final Sheets sheets = workbook.getSheets();
	    final List<Sheet> sheetList = sheets.getSheet();
	    for(int sheetIndex = 0; sheetIndex < sheetList.size(); sheetIndex++) {
	    	final Sheet sheet = sheetList.get(sheetIndex);
	    	final Part part = getSheetPart(sheet);
	    	final String sheetType = getSheetType(part);
	        createInsertSheetOperation(sheet, sheetType, part, sheetIndex);
	        if(sheetType.equals("worksheet")) {
	        	if(activeSheetIndex==sheetIndex) {
	        		createWorksheetOperations((WorksheetPart)part, sheetIndex, true);
	        	}
	        }
	    }
	    return activeSheetIndex;
    }

    public void createOutstandingOperations(int activeSheetIndex)
        throws Exception {

    	final WorkbookPart workbookPart = spreadsheetMLPackage.getWorkbookPart();
	    final Workbook workbook = workbookPart.getJaxbElement();

		final Sheets sheets = workbook.getSheets();
	    final List<Sheet> sheetList = sheets.getSheet();
	    for(int sheetIndex = 0; sheetIndex < sheetList.size(); sheetIndex++) {
	    	final Sheet sheet = sheetList.get(sheetIndex);
	    	final Part part = getSheetPart(sheet);
	    	final String sheetType = getSheetType(part);
	        if(sheetType.equals("worksheet")) {
	        	if(activeSheetIndex!=sheetIndex) {
	        		createWorksheetOperations((WorksheetPart)part, sheetIndex, false);
	        	}
	        }
	    }
    }

    private void checkComplexity()
        throws FilterException {

        final WorkbookPart workbookPart = spreadsheetMLPackage.getWorkbookPart();
        final Workbook workbook = workbookPart.getJaxbElement();
        final ServiceLookup services = getOperationDocument().getServiceLookup();
        final Session session = getOperationDocument().getSession();

        int maxCellCount  =      ConfigurationHelper.getIntegerOfficeConfigurationValue(services, session, "//spreadsheet/maxCells", 500000);
        int maxSheetCount =      ConfigurationHelper.getIntegerOfficeConfigurationValue(services, session, "//spreadsheet/maxSheets", 256);

        final Sheets sheets = workbook.getSheets();
        if(sheets!=null) {
            final List<Sheet> sheetList = sheets.getSheet();
            if(maxSheetCount>=0) {
                if(sheetList.size()>maxSheetCount) {
                    throw new FilterException("", ErrorCode.COMPLEXITY_TOO_HIGH);
                }
            }
            if(maxCellCount>0) {
                final RelationshipsPart relationshipPart = spreadsheetMLPackage.getWorkbookPart().getRelationshipsPart();
                for(Sheet sheet:sheetList) {
                    final Part part = relationshipPart.getPart(sheet.getId());
                    if(part instanceof WorksheetPart) {
                        final Worksheet worksheet = ((WorksheetPart)part).getJaxbElement();
                        final SheetData sheetData = worksheet.getSheetData();
                        if(sheetData!=null) {
                            final Iterator<Row> rowIterator = sheetData.createRowIterator();
                            while(rowIterator.hasNext()) {
                                final Row row = rowIterator.next();
                                maxCellCount -= row.getCellCount();
                                if(maxCellCount<0) {
                                    throw new FilterException("", ErrorCode.COMPLEXITY_TOO_HIGH);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /* returns the active sheet */

    private int createDocumentSettings(Workbook workbook)
    	throws JSONException {

    	long activeSheet = 0;
		final JSONObject documentAttrs = new JSONObject();
		documentAttrs.put("fileFormat", "ooxml");
		documentAttrs.put("cols", 16384);
		documentAttrs.put("rows", 1048576);
		final BookViews bookViews = workbook.getBookViews();
    	if(bookViews!=null) {
    		final List<CTBookView> bookViewList = bookViews.getWorkbookView();
    		if(!bookViewList.isEmpty()) {
    			final CTBookView bookView = bookViewList.get(0);
    			activeSheet = bookView.getActiveTab();
    			documentAttrs.put("activeSheet", activeSheet);
    		}
    	}
		final JSONObject attrs = new JSONObject(1);
		attrs.put("document", documentAttrs);
		addSetDocumentAttributesOperation(attrs);

		return (int)activeSheet;
    }

    private void createInsertNameOperations(Workbook workbook)
        throws JSONException {

        final DefinedNames definedNames = workbook.getDefinedNames();
        if(definedNames==null) {
            return;
        }
        final List<CTDefinedName> definedNameList = definedNames.getDefinedName();
        for(CTDefinedName definedName:definedNameList) {
            addInsertNameOperation(definedName.getLocalSheetId(), definedName.getName(), definedName.getValue()!=null ? definedName.getValue() : "");
        }
    }

    private JSONObject getCellObject(Cell cell, boolean preview)
        throws Exception {

    	Object value = null;
    	Object formula = null;

        final JSONObject cellObject = new JSONObject();
        final CTCellFormula cellFormula = cell.getF();
        final boolean isFormula = cellFormula!=null&&(cellFormula.getT()==STCellFormulaType.NORMAL||cellFormula.getT()==STCellFormulaType.SHARED);
        if(isFormula) {
        	final boolean hasFormulaText = cellFormula.getValue()!=null&&cellFormula.getValue().length()>0;
            if(cellFormula.getT()==STCellFormulaType.SHARED) {
                final Long si = cellFormula.getSi();
                if(si!=null) {
                    cellObject.put("shared", si+1);
                }
                if(hasFormulaText) {
                    final SmlUtils.CellRef cellRef = new SmlUtils.CellRef(cell.getColumn(), ((Row)cell.getParent()).getRow());
                    cellObject.put("ref", cellRef.getJSONArray());
                }
            }
            if(hasFormulaText)
                formula = new String("=" + cellFormula.getValue());
        }
        if(preview||!isFormula) {
        	if (cell.getT()==STCellType.INLINE_STR) {
	            CTRst is = cell.getIs();
	            if(is!=null) {
	                // TODO: CTRst can include formatted text runs, but we will support simple t first.
	                if(is.getT()!=null) {
	                	value = new String(is.getT().getValue());
	                }
	            }
	        }
	        else if(cell.getV()!=null) {
	            if(cell.getT()==STCellType.N) {
	                try {
	                    value = new Double(Double.parseDouble(cell.getV()));
	                }
	                catch(NumberFormatException e) {
	                    // TODO: ups
	                }
	            }
	            else if(cell.getT()==STCellType.S) {
	                String sharedString = getOperationDocument().getSharedString(cell.getV());
	                if(sharedString!=null) {
	                    value = new String(getOperationDocument().escapeString(sharedString));
	                }
	            }
	            else if(cell.getT()==STCellType.B) {
	                if(cell.getV().equals("0")||cell.getV().toLowerCase().equals("false")) {
	                	value = new Boolean(false);
	                }
	                else if(cell.getV().equals("1")||cell.getV().toLowerCase().equals("true")) {
	                    value = new Boolean(true);
	                }
	            }
	            else {  // STCellType.STR, STCellType.E or STCellType.F (ARRAY or DATA_TABLE format)
	            	final String cellValue = cell.getV();
	                value = new String(cellValue.startsWith("=") ? "'" + cellValue : cellValue);
	            }
	        }
        }
    	if(isFormula) {
    		if(formula!=null) {
    			cellObject.put("value", formula);
    		}
        	if(value!=null) {
        		cellObject.put("result", value);
        	}
    	}
    	else if (value!=null) {
    		cellObject.put("value", value);
    	}
        final JSONObject jsonCellAttrs = CellUtils.createCellProperties(cell, getOperationDocument().getStylesheet(true));
        if(jsonCellAttrs!=null) {
           cellObject.put("attrs", jsonCellAttrs );
        }
        return cellObject;
    }

    private Part getSheetPart(final Sheet sheet) {
        final RelationshipsPart relationshipPart = spreadsheetMLPackage.getWorkbookPart().getRelationshipsPart();
        return relationshipPart.getPart(sheet.getId());
    }

    private String getSheetType(final Part part)
    	throws FilterException {

        String sheetType = null;
        if(part.getRelationshipType().equals("http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet")) {
            sheetType = "worksheet";
        }
        else if(part.getRelationshipType().equals("http://schemas.microsoft.com/office/2006/relationships/xlMacrosheet")) {
            sheetType = "macrosheet";
        }
        else if(part.getRelationshipType().equals("http://schemas.microsoft.com/office/2006/relationships/xlIntlMacrosheet")) {
            sheetType = "macrosheet";
        }
        else if(part.getRelationshipType().equals("http://schemas.openxmlformats.org/officeDocument/2006/relationships/dialogsheet")) {
            sheetType = "dialogsheet";
        }
        else if(part.getRelationshipType().equals("http://schemas.openxmlformats.org/officeDocument/2006/relationships/chartsheet")) {
            sheetType = "chartsheet";
        }
        if(sheetType==null) {
            throw new FilterException("xlsx import: unknown sheettype detected.", ErrorCode.CRITICAL_ERROR);
        }
        return sheetType;
    }

    /*
     * if != null the topLeftCell parameter is filled with the current sheet settings
     */
    private void createInsertSheetOperation(Sheet sheet, String sheetType, Part part, int sheetIndex)
    	throws JSONException {

        if(sheetType.equals("worksheet")) {
            final WorksheetPart worksheetPart = (WorksheetPart)part;
            final Worksheet worksheet = worksheetPart.getJaxbElement();

            final JSONObject properties = new JSONObject();
            properties.put("sheet", SheetUtils.createWorksheetProperties(sheet, worksheet));
            properties.put("column", SheetUtils.createWorksheetColumnProperties(getOperationDocument(), worksheet));
            properties.put("row",  SheetUtils.createWorksheetRowProperties(worksheet));
            addInsertSheetOperation(sheetIndex, sheet.getName(), sheetType, properties);
        }
        else {
            addInsertSheetOperation(sheetIndex, sheet.getName(), sheetType, Commons.surroundJSONObject("sheet", SheetUtils.createSheetProperties(sheet)));
        }
    }

    private void createWorksheetOperations(final WorksheetPart worksheetPart, int sheetIndex, boolean preview)
        throws Exception {

    	final Worksheet worksheet = worksheetPart.getJaxbElement();

        // column attributes... row attributes ... cells...
        createColumnOperations(worksheet.getCols(), sheetIndex);
        createRowOperations(worksheet, sheetIndex, preview);

        MergeCellHelper.createMergeCellOperations(operationsArray, sheetIndex, worksheet.getMergeCells());

        // hyperlinks
        final CTHyperlinks hyperlinks = worksheet.getHyperlinks();
        if(hyperlinks!=null) {
            final List<CTHyperlink> hyperlinkList = hyperlinks.getHyperlink();
            for(int i=0; i<hyperlinkList.size();i++) {
                final CTHyperlink hyperlink = hyperlinkList.get(i);
                final SmlUtils.CellRef cellRef = SmlUtils.createCellRef(hyperlink.getRef());
                if(cellRef!=null) {
                    final String hyperlinkUrl = Commons.getUrl(worksheetPart, hyperlink.getId());
                    if(hyperlinkUrl!=null&&!hyperlinkUrl.isEmpty()) {
                        final JSONObject attrs = new JSONObject();
                        final JSONObject characterAttributes = new JSONObject();
                        characterAttributes.put("url", hyperlinkUrl);
                        attrs.put("character", characterAttributes);
                        addSetCellAttributesOperation(sheetIndex, cellRef, attrs);
                    }
                }
            }
        }
        //drawings
        final CTDrawing drawing = worksheet.getDrawing();
        if(drawing!=null) {
            final String drawingId = drawing.getId();
            if(drawingId!=null&&drawingId.length()>0) {
                final RelationshipsPart worksheetRelationships = worksheetPart.getRelationshipsPart();
                if(worksheetRelationships!=null) {
                    final Part drawingsPart = worksheetRelationships.getPart(drawingId);
                    if(drawingsPart.getRelationshipType().equals("http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing")&&drawingsPart instanceof org.docx4j.openpackaging.parts.DrawingML.Drawing) {
                        Drawings.createDrawingOperations(operationsArray, (org.docx4j.openpackaging.parts.DrawingML.Drawing)drawingsPart, sheetIndex);
                    }
                }
            }
        }

        // validations
        final List<IDataValidations> validations = getOperationDocument().getDataValidations(worksheet);
        for(IDataValidations dataValidations:validations) {
            for(IDataValidation dataValidation:dataValidations.getDataValidation()) {
                createDataValidation(sheetIndex, dataValidation);
            }
        }
    }

    private void createDataValidation(int sheetIndex, IDataValidation dataValidation)
        throws JSONException {

        final List<SmlUtils.CellRefRange> cellRefRangeList = new ArrayList<SmlUtils.CellRefRange>();
        final List<String> sqRefList = dataValidation.getsqref();
        for(String sqRef:sqRefList) {
            final SmlUtils.CellRefRange cellRefRange = SmlUtils.createCellRefRange(sqRef);
            if(cellRefRange!=null) {
                cellRefRangeList.add(cellRefRange);
            }
        }
        if(cellRefRangeList.size()>0) {
            ValidationUtils.addInsertValidationOperation(operationsArray, sheetIndex, cellRefRangeList, dataValidation.getType(), dataValidation.getOperator(), dataValidation.getFormula1(), dataValidation.getFormula2(), dataValidation.isShowInputMessage(),
                dataValidation.getPromptTitle(), dataValidation.getPrompt(), dataValidation.isShowErrorMessage(), dataValidation.getErrorTitle(), dataValidation.getError(), dataValidation.getErrorStyle(),
                    dataValidation.isShowDropDown()==false, dataValidation.isAllowBlank());
        }
    }

    private boolean compareCell(Cell source, Cell dest) {
    	if(source.getF()!=null||dest.getF()!=null)
    		return false;
    	if(source.getIs()!=null||dest.getIs()!=null)
    		return false;
    	if(source.getT()!=dest.getT())
    		return false;
    	if(source.getVm()!=dest.getVm())
    		return false;
    	if(source.getCm()!=dest.getCm())
    		return false;
    	if(source.getS()!=dest.getS())
    		return false;
    	if(source.getV()==null&&dest.getV()==null)
    		return true;
    	if(source.getV()!=null&&dest.getV()!=null) {
    		return source.getV().equals(dest.getV());
    	}
    	else {
    		return false;
    	}
    }

    private void createRowOperations(Worksheet worksheet, int sheetIndex, boolean preview)
        throws Exception {
    	
    	final SheetData sheetData = worksheet.getSheetData();
    	if(sheetData==null) {
    		return;
    	}

/*
    	// about 169 bytes required for an empty row
    	// now 144 bytes
    	getOperationDocument().getFreeHeapSpace();
    	for(int i=100000; i<600000; i++) {
    		sheetData.getRow(i, true);
    	}

    	// about 139 bytes required for an empty cell
    	// now 131 bytes
    	getOperationDocument().getFreeHeapSpace();
    	for(int j = 100000; j<101000; j++) {
    		Row asd = sheetData.getRow(j, true);
    		for(int k=0; k<500; k++) {
    			Cell blah = asd.getCell(k, true);
    		}
    	}
    	getOperationDocument().getFreeHeapSpace();
*/
    	boolean sheetRowsHiddenByDefault = false;
    	final CTSheetFormatPr sheetFormatPr = worksheet.getSheetFormatPr();
    	if(sheetFormatPr!=null) {
    		if(sheetFormatPr.isZeroHeight()) {
    			sheetRowsHiddenByDefault = true;
    		}
    	}
    	// first we will create row attributes
        Iterator<Row> rowIterator = sheetData.createRowIterator();

        Row lastRow = null;
        int lastRowNumber = 0;
        int rowCount = 0;

        while(rowIterator.hasNext()) {
            final Row row = rowIterator.next();

            if(lastRow!=null&&row.getRow()==lastRowNumber+rowCount&&RowUtils.compareRowProperties(lastRow, row)) {
                rowCount++;
            }
            else {
                if(lastRow!=null) {
                    final JSONObject attrs = RowUtils.createRowProperties(lastRow, getOperationDocument().getStylesheet(true), sheetRowsHiddenByDefault);
                    if(attrs!=null) {
                        addSetRowAttributes(sheetIndex, lastRowNumber, (lastRowNumber - 1) + rowCount, attrs);
                    }
                }
                lastRow = row;
                lastRowNumber = row.getRow();
                rowCount = 1;
            }
            if(!rowIterator.hasNext()&&lastRow!=null) {
                final JSONObject attrs = RowUtils.createRowProperties(lastRow, getOperationDocument().getStylesheet(true), sheetRowsHiddenByDefault);
                if(attrs!=null) {
                    addSetRowAttributes(sheetIndex, lastRowNumber, (lastRowNumber - 1) + rowCount, attrs);
                }
            }
        }

        // create cell content for each row
        rowIterator = sheetData.createRowIterator();
        while(rowIterator.hasNext()) {
        	final Row row = rowIterator.next();

        	final JSONArray rowArray = new JSONArray();
            final JSONArray cellArray = new JSONArray();
            SmlUtils.CellRef topLeftCellRef = null;

            Cell 		lastCell = null;
            JSONObject 	lastCellObject = null;
            int 		lastCellX = -1;

            Iterator<Cell> cellIterator = row.createCellIterator();
            while(cellIterator.hasNext()) {
            	boolean repeat = false;
                Cell cell = cellIterator.next();
                final SmlUtils.CellRef cellRef = new SmlUtils.CellRef(cell.getColumn(), row.getRow());
                if(topLeftCellRef==null) {
                    topLeftCellRef = cellRef;
                }
                if(lastCell!=null) {
	                // checking empty cells
	                final int cellDifference = cellRef.getColumn() - lastCellX;
	                if(cellDifference > 1) {
	                	if(lastCellObject!=null) {
	                		cellArray.put(lastCellObject);
	                	}
	                	// we have empty cells
	                	if(cellDifference==2) {
	                		cellArray.put((Object)null);
	                	}
	                	else {
	                		final JSONObject emptyCells = new JSONObject(1);
	                		emptyCells.put("repeat", cellDifference - 1);
	                		cellArray.put(emptyCells);
	                	}
	                }
	                else {
	                	if(compareCell(lastCell, cell)) {
	                		lastCellObject.put("repeat", lastCellObject.optInt("repeat", 1) + 1);
	                		repeat = true;
	                	}
	                	else {
	                		cellArray.put(lastCellObject);
	                	}
	                }
                }
            	lastCell = cell;
             	lastCellX = cellRef.getColumn();
            	if(repeat==false) {
	            	lastCellObject = getCellObject(cell, preview);
            	}
            }
            if(lastCellObject!=null) {
                cellArray.put(lastCellObject);
            }
            rowArray.put(cellArray);
            if(topLeftCellRef!=null)
                addSetCellContentsOperation(sheetIndex, topLeftCellRef, rowArray);
        }
    }

    private void createColumnOperations(List<Cols> colsIter, int sheetIndex)
        throws Exception {

        if(colsIter==null)
            return;

        for(Cols cols:colsIter) {
            List<Col> colIter = cols.getCol();
            if(colIter!=null) {
                for(Col col:colIter) {
                    JSONObject attrs = ColumnUtils.createColumnProperties(getOperationDocument(), col);
                    if(attrs!=null) {
                        addSetColumnAttributes(sheetIndex, col.getMin()-1, col.getMax()-1, attrs);
                    }
                }
            }
        }
    }

    @Override
	public void addSetDocumentAttributesOperation(JSONObject attrs)
    	throws JSONException {

    	if(attrs!=null&&!attrs.isEmpty()) {
	    	final JSONObject setDocumentAttributesObject = new JSONObject(2);
	    	setDocumentAttributesObject.put("name", "setDocumentAttributes");
	    	setDocumentAttributesObject.put("attrs", attrs);
	    	operationsArray.put(setDocumentAttributesObject);
    	}
    }

    public void addInsertSheetOperation(int sheetIndex, final String sheetName, final String type, final JSONObject attrs)
        throws JSONException {

        final JSONObject insertSheetObject = new JSONObject();
        insertSheetObject.put("name", "insertSheet");
        insertSheetObject.put("sheet", sheetIndex);
        if(type!=null&&!type.equals("worksheet")) {
            insertSheetObject.put("type", type);
        }
        insertSheetObject.put("sheetName", sheetName);
        if(attrs!=null)
            insertSheetObject.put("attrs", attrs);

        operationsArray.put(insertSheetObject);
    }

    final static List<Object> objectList = new ArrayList<Object>();

    public void addSetCellContentsOperation(int sheetIndex, SmlUtils.CellRef cellRef, JSONArray cellContent)
        throws JSONException {

        final JSONObject insertCellsObject = new JSONObject(4);
        insertCellsObject.put("name", "setCellContents");
        insertCellsObject.put("sheet", sheetIndex);
        insertCellsObject.put("start", cellRef.getJSONArray());
        insertCellsObject.put("contents", cellContent);

        operationsArray.put(insertCellsObject);
    }

    public void addSetCellAttributesOperation(int sheetIndex, SmlUtils.CellRef cellRef, JSONObject attrs)
        throws Exception {

        if(attrs!=null&&!attrs.isEmpty()) {
            final JSONObject fillCellRangeObject = new JSONObject(4);
            fillCellRangeObject.put("name", "fillCellRange");
            fillCellRangeObject.put("sheet", sheetIndex);
            fillCellRangeObject.put("start", cellRef.getJSONArray());
            fillCellRangeObject.put("attrs", attrs);

            operationsArray.put(fillCellRangeObject);
        }
    }

    public void addSetRowAttributes(int sheetIndex, int start, int end, JSONObject attrs)
        throws JSONException {

        if(attrs==null||attrs.length()==0)
            return;

        final JSONObject addSetRowAttributesObject = new JSONObject();
        addSetRowAttributesObject.put("name", "setRowAttributes");
        addSetRowAttributesObject.put("sheet", sheetIndex);
        addSetRowAttributesObject.put("start", start);
        if(end!=start) {
            addSetRowAttributesObject.put("end", end);
        }
        if(attrs!=null)	{
        	addSetRowAttributesObject.put("attrs", attrs);
        }
        operationsArray.put(addSetRowAttributesObject);
    }

    public void addSetColumnAttributes(int sheetIndex, long start, long end, JSONObject attrs)
        throws JSONException {

        if(attrs==null||attrs.length()==0)
            return;

        final JSONObject addSetColumnAttributesObject = new JSONObject();
        addSetColumnAttributesObject.put("name", "setColumnAttributes");
        addSetColumnAttributesObject.put("sheet", sheetIndex);
        addSetColumnAttributesObject.put("start", start);
        if(end>start) {
            addSetColumnAttributesObject.put("end", end);
        }
        addSetColumnAttributesObject.put("attrs", attrs);

        operationsArray.put(addSetColumnAttributesObject);
    }

    public void addInsertNameOperation(Long sheetIndex, String name, String formula)
        throws JSONException {

        final JSONObject addInsertNameObject = new JSONObject();
        addInsertNameObject.put("name", "insertName");
        if(sheetIndex!=null) {
            addInsertNameObject.put("sheet", sheetIndex);
        }
        addInsertNameObject.put("exprName", name);
        addInsertNameObject.put("formula", formula);

        operationsArray.put(addInsertNameObject);
    }

}
