package com.openexchange.office.tools.htmldoc;

import org.json.JSONArray;
import org.json.JSONObject;

public class Table
		extends NodeHolder
		implements ISplitpoint
{
	private final static String	TABLE_SIZEEXCEEDED_PLACEHOLDER_START       = "<table role=\"grid\" class=\"size-exceeded\" contenteditable=\"false\" size-exceed-values=";
    private final static String TABLE_SIZEEXCEEDED_PLACEHOLDER_END         = "><colgroup></table>";
	private final static String	TABLE_SIZEEXCEEDED_PLACEHOLDER_SPLIT_START = "<table role=\"grid\" class=\"size-exceeded splitpoint\" contenteditable=\"false\" size-exceed-values=";

	private final TextTableLimits tableLimits;
	private final JSONArray       tableGrid;
	private final int             allW;
	private int                   cellCount = 0;
	private int                   rowCount = 0;
	private int                   columnCount = 0;
	private boolean               splitpoint = false;
	private JSONObject sizeExceeded;

	public Table(
			final TextTableLimits textTableLimits,
			JSONObject attrs, 
			JSONObject sizeExceeded)
		throws Exception
	{
		
		this.sizeExceeded = sizeExceeded;		

		final JSONObject table = attrs.getJSONObject("table");

		this.tableGrid = table.getJSONArray("tableGrid");
		this.tableLimits = textTableLimits;

		
		int allW = 0;
		for (int i = 0; i < tableGrid.length(); i++)
		{
			allW += tableGrid.getInt(i);
		}
		this.allW = allW;
		columnCount = tableGrid.length();
	}

	@Override
	public void insert(
		JSONArray start,
		INode newChild)
		throws Exception
	{

		if (start.length() == 1 && newChild instanceof Row)
		{
			((Row) newChild).setCellCount(tableGrid.length());
			rowCount++;
			cellCount += columnCount;
		}
		else if (newChild instanceof Paragraph)
		{
			((Paragraph) newChild).setInTable(true);
		}

		super.insert(start, newChild);
	}

	@Override
	public boolean appendContent(
		StringBuilder document)
		throws Exception
	{
		if(sizeExceeded == null || sizeExceeded.isEmpty())
        {
			document.append("<table");
			if (splitpoint) {
				document.append(" class=\"splitpoint\" ");
			}
			GenDocHelper.appendAttributes(getAttribute(), document);
			document.append("role='grid' style='width: 100%; background-color: transparent;'><colgroup>");
			for (int i = 0; i < tableGrid.length(); i++)
			{

				final double w = Math.round((tableGrid.getDouble(i) / allW) * 1000.0) / 10.0;

				document.append("<col style='width: " + w + "%;'>");
			}
			document.append("</colgroup><tbody>");

			super.appendContent(document);

			document.append("</tbody></table>");
        }
        else
        {
        	if (splitpoint)
        		document.append(TABLE_SIZEEXCEEDED_PLACEHOLDER_SPLIT_START);
        	else
        	    document.append(TABLE_SIZEEXCEEDED_PLACEHOLDER_START);
        	
        	if(sizeExceeded.has("rows")){
        		rowCount = sizeExceeded.getInt("rows");
        	}
        	if(sizeExceeded.has("columns")){
        		columnCount = sizeExceeded.getInt("columns");
        	}
        	
        	if (rowCount > tableLimits.maxTableRows) {
        		document.append("\"rows," + rowCount + "," + tableLimits.maxTableRows + "\"");
        	} else if (columnCount > tableLimits.maxTableColumns) {
        		document.append("\"cols," + columnCount + "," + tableLimits.maxTableColumns + "\"");
        	} else if (cellCount < tableLimits.maxTableCells) {
        		document.append("\"cells," + cellCount + "," + tableLimits.maxTableCells + "\"");
        	}
        	document.append(TABLE_SIZEEXCEEDED_PLACEHOLDER_END);
        }
		return true;
	}

	@Override
	public boolean needsEmptySpan()
	{
		return false;
	}

	@Override
	public boolean isSplitpoint() {
		return splitpoint;
	}

	@Override
	public void setSplitpoint(boolean splitpoint) {
		this.splitpoint = splitpoint;
    }

	public int getRowCount() {
	    return rowCount;
	}
}
